/* -*-js-*-
 *
 * Common JavaScript code for Chopwood
 *
 * (c) 2013 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Chopwood: a password-changing service.
 *
 * Chopwood is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * Chopwood is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with Chopwood; if not, see
 * <http://www.gnu.org/licenses/>.
 */

/*----- Some utilities ----------------------------------------------------*/

function elt(id) {
  /* Return the element with the requested ID. */
  return document.getElementById(id);
}

function map(func, list) {
  /* Apply FUNC to each element of LIST, which may actually be any object;
   * return a new object mapping the same keys to the images of the values in
   * the function FUNC.
   */
  var i;
  var out = {};
  for (i in list) out[i] = func(list[i]);
  return out;
}

/*----- Form validation ---------------------------------------------------*/

var FORMS = {};
/* A map of form names to information about them.  Each form is an object
 * with the following slots:
 *
 *   * elts: A list of element-ids for the form widgets.  These widgets will
 *     be checked periodically to see whether the input data is acceptable.
 *
 *   * check: A function of no arguments, which returns either `null' if
 *     everything is OK, or an error message as a string.
 *
 * Form names aren't just for show.  Some element-ids are constructed using
 * the form name as a base:
 *
 *   * FORM-whinge: An output element in which to display an error message if
 *     the form's input is unacceptable.
 *
 *   * FORM-submit: The Submit button, which needs hooking to inhibit
 *     submitting a form with invalid data.
 */

function check() {
  /* Check through the various forms to make sure they're filled in OK.  If
    * not, set the `F-whinge' elements, and disable `F-submit'.
    */
  var f, form, whinge;

  for (f in FORMS) {
    form = FORMS[f];
    we = elt(f + '-whinge');
    sb = elt(f + '-submit');
    whinge = form.check();
    if (sb !== null) sb.disabled = (whinge !== null);
    if (we !== null) {
      we.textContent = whinge || 'OK';
      we.className = whinge === null ? 'whinge' : 'whinge wrong';
    }
  }

  // We can't catch all possible change events: in particular, it seems
  // really hard to capture changes as a result of selections from a menu --
  // e.g., delete or paste.  Accept this, and just recheck periodically.
  check_again(1000);
}

var timer = null;
/* The timer for the periodic validation job. */

function check_again(when) {
  /* Arrange to check the forms again in WHEN milliseconds. */
  if (timer !== null) clearTimeout(timer);
  timer = setTimeout(check, when);
}

var Q = 0;
function check_soon(ev) {
  /* Arrange to check the forms again very soon. */
  check_again(50);
}

function check_presubmit(ev, f) {
  /* Check the form F now, popping up an alert and preventing the event EV if
   * there's something wrong.
   */
  var whinge = FORMS[f].check();

  if (whinge !== null) {
    ev.preventDefault();
    alert(whinge);
  }
}

function init() {
  /* Attach event handlers to the various widgets so that we can keep track
   * of how well things are being filled in.
   */
  var f, form, w, e;

  // Start watching for changes.
  check_soon();

  for (f in FORMS) (function (f, form) {

    // Ugh.  We have to lambda-bind `f' here so that we can close over it
    // properly.
    for (w in form.elts) {
      if ((e = elt(f + '-' + form.elts[w])) === null) continue;
      e.addEventListener('click', check_soon);
      e.addEventListener('change', check_soon);
      e.addEventListener('keypress', check_soon);
      e.addEventListener('blur', check_soon);
    }
    if ((e = elt(f + '-submit')) !== null) {
      e.addEventListener('click', function (ev) {
	return check_presubmit(ev, f)
      });
    }
  })(f, FORMS[f]);
}

window.addEventListener('load', init);

/*----- That's all, folks -------------------------------------------------*/
